@@ -2,7 +2,7 @@
  * @file        https://github.com/Zilong-Li/vcfpp/vcfpp.h
  * @author      Zilong Li
  * @email       zilong.dk@gmail.com
- * @version     v0.3.4
+ * @version     v0.3.8
  * @breif       a single C++ file for manipulating VCF
  * Copyright (C) 2022-2023.The use of this code is governed by the LICENSE file.
  ******************************************************************************/
@@ -118,6 +118,15 @@ inline bool isEndWith(std::string const & s, std::string const & e)
     }
 }
 
+// determinate the mode for read/write the compressed/uncompressed VCF/BCF
+inline std::string getMode(std::string const & fname, std::string mode = "r")
+{
+    if(isEndWith(fname, "bcf.gz")) mode += "b";
+    if(isEndWith(fname, "bcf")) mode += "bu";
+    if(isEndWith(fname, "vcf.gz")) mode += "z";
+    return mode;
+}
+
 // string split by separator
 inline std::vector<std::string> split_string(const std::string & s, const std::string & separators)
 {
@@ -157,6 +166,15 @@ struct variant_close
     }
 };
 
+// deleter for bcf header
+struct bcf_hdr_close
+{
+    void operator()(bcf_hdr_t * x)
+    {
+        if(x) bcf_hdr_destroy(x);
+    }
+};
+
 /**
  * @class BcfHeader
  * @brief Object represents header of the VCF, offering methods to access and modify the tags
@@ -177,8 +195,8 @@ class BcfHeader
 
     ~BcfHeader()
     {
-        if(hdr) bcf_hrec_destroy(hrec);
-        if(hrec) bcf_hdr_destroy(hdr);
+        if(hrec) bcf_hrec_destroy(hrec);
+        if(hdr) bcf_hdr_destroy(hdr);
     }
 
     /** @brief stream out the header */
@@ -384,9 +402,9 @@ class BcfRecord
     friend class BcfWriter;
 
   private:
-    BcfHeader header;
+    BcfHeader * header;
     std::shared_ptr<bcf1_t> line = std::shared_ptr<bcf1_t>(bcf_init(), variant_close()); // variant
-    bcf_hdr_t * hdr_d; // a dup header by bcf_hdr_dup(header->hdr)
+    bcf_hdr_t * hdr_d = NULL; // a dup header by bcf_hdr_dup(header->hdr)
     bcf_fmt_t * fmt = NULL;
     bcf_info_t * info = NULL;
     int32_t * gts = NULL;
@@ -405,23 +423,33 @@ class BcfRecord
     BcfRecord() {}
 
     /// constructor with a given BcfHeader object
-    BcfRecord(const BcfHeader & h)
+    BcfRecord(BcfHeader & h)
     {
-        init(h);
+        initHeader(h);
     }
 
-    ~BcfRecord() {}
-
-    /// initilize a BcfRecord object using a given BcfHeader object
-    void init(const BcfHeader & h)
+    ~BcfRecord()
     {
-        if(h.hdr == NULL) throw std::runtime_error("please initial header first\n");
-        header = h;
-        nsamples = header.nSamples();
+        if(gts) free(gts);
+        if(hdr_d) bcf_hdr_destroy(hdr_d);
+    }
+
+    /// initilize the header associated with BcfRecord object by pointing to another BcfHeader object
+    void initHeader(BcfHeader & h)
+    {
+        header = &h;
+        if(!header->hdr) throw std::runtime_error("please initial header first\n");
+        nsamples = header->nSamples();
         typeOfGT.resize(nsamples);
         gtPhase.resize(nsamples, 0);
     }
 
+    /// reset the header associated with BcfRecord object by pointing to another BcfHeader object
+    void resetHeader(BcfHeader & h)
+    {
+        header = &h;
+    }
+
     /** @brief stream out the variant */
     friend std::ostream & operator<<(std::ostream & out, const BcfRecord & v)
     {
@@ -433,7 +461,7 @@ class BcfRecord
     inline std::string asString() const
     {
         kstring_t s = {0, 0, NULL}; // kstring
-        if(vcf_format(header.hdr, line.get(), &s) == 0)
+        if(vcf_format(header->hdr, line.get(), &s) == 0)
         {
             std::string out(s.s, s.l);
             free(s.s);
@@ -455,7 +483,7 @@ class BcfRecord
     isValidGT<T> getGenotypes(T & v)
     {
         ndst = 0;
-        ret = bcf_get_genotypes(header.hdr, line.get(), &gts, &ndst);
+        ret = bcf_get_genotypes(header->hdr, line.get(), &gts, &ndst);
         if(ret <= 0) throw std::runtime_error("genotypes not present");
         // if nploidy is not set manually. find the max nploidy using the first variant (eg. 2) resize v as
         // max(nploidy)
@@ -475,7 +503,7 @@ class BcfRecord
         isGenoMissing.assign(nsamples, 0);
         int i, j, nphased = 0;
         noneMissing = true;
-        fmt = bcf_get_fmt(header.hdr, line.get(), "GT");
+        fmt = bcf_get_fmt(header->hdr, line.get(), "GT");
         int nploidy_cur = ret / nsamples; // requires nploidy_cur <= nploidy
         for(i = 0; i < nsamples; i++)
         {
@@ -518,7 +546,7 @@ class BcfRecord
     bool getGenotypes(std::vector<int> & v)
     {
         ndst = 0;
-        ret = bcf_get_genotypes(header.hdr, line.get(), &gts, &ndst);
+        ret = bcf_get_genotypes(header->hdr, line.get(), &gts, &ndst);
         if(ret <= 0)
             throw std::runtime_error(
                 "genotypes not present. make sure you initilized the variant object first\n");
@@ -568,18 +596,18 @@ class BcfRecord
     template<typename T, typename S = typename T::value_type>
     isFormatVector<T> getFORMAT(std::string tag, T & v)
     {
-        fmt = bcf_get_fmt(header.hdr, line.get(), tag.c_str());
+        fmt = bcf_get_fmt(header->hdr, line.get(), tag.c_str());
         if(!fmt) throw std::runtime_error("there is no " + tag + " in FORMAT of this variant.\n");
         nvalues = fmt->n;
         ndst = 0;
         S * dst = NULL;
-        int tagid = header.getFormatType(tag);
+        int tagid = header->getFormatType(tag);
         if(tagid == 1)
-            ret = bcf_get_format_int32(header.hdr, line.get(), tag.c_str(), &dst, &ndst);
+            ret = bcf_get_format_int32(header->hdr, line.get(), tag.c_str(), &dst, &ndst);
         else if(tagid == 2)
-            ret = bcf_get_format_float(header.hdr, line.get(), tag.c_str(), &dst, &ndst);
+            ret = bcf_get_format_float(header->hdr, line.get(), tag.c_str(), &dst, &ndst);
         else if(tagid == 3)
-            ret = bcf_get_format_char(header.hdr, line.get(), tag.c_str(), &dst, &ndst);
+            ret = bcf_get_format_char(header->hdr, line.get(), tag.c_str(), &dst, &ndst);
         else
             throw std::runtime_error("can not find the type of " + tag + " in the header file.\n");
         if(ret >= 0)
@@ -591,6 +619,7 @@ class BcfRecord
         }
         else
         {
+            free(dst);
             throw std::runtime_error("failed to parse the " + tag + " format of this variant.\n");
         }
     }
@@ -603,14 +632,14 @@ class BcfRecord
      * */
     bool getFORMAT(std::string tag, std::vector<std::string> & v)
     {
-        fmt = bcf_get_fmt(header.hdr, line.get(), tag.c_str());
+        fmt = bcf_get_fmt(header->hdr, line.get(), tag.c_str());
         if(!fmt) throw std::runtime_error("there is no " + tag + " in FORMAT for this variant of ID=" + ID());
         nvalues = fmt->n;
         // if ndst < (fmt->n+1)*nsmpl; then realloc is involved
         ret = -1, ndst = 0;
         char ** dst = NULL;
-        if(header.getFormatType(tag) == 3)
-            ret = bcf_get_format_string(header.hdr, line.get(), tag.c_str(), &dst, &ndst);
+        if(header->getFormatType(tag) == 3)
+            ret = bcf_get_format_string(header->hdr, line.get(), tag.c_str(), &dst, &ndst);
         if(ret > 0)
         {
             v.clear();
@@ -619,11 +648,12 @@ class BcfRecord
                 // Ugly: GT field is considered to be a string by the VCF header but BCF represents it as INT.
                 v.emplace_back(dst[i]);
             };
-            free(dst);
+            free(dst[0]); free(dst);
             return true;
         }
         else
         {
+            free(dst[0]); free(dst);
             throw std::runtime_error("couldn't parse the " + tag + " format of this variant.\n");
         }
     }
@@ -637,24 +667,29 @@ class BcfRecord
     template<typename T, typename S = typename T::value_type>
     isInfoVector<T> getINFO(std::string tag, T & v)
     {
-        info = bcf_get_info(header.hdr, line.get(), tag.c_str());
+        info = bcf_get_info(header->hdr, line.get(), tag.c_str());
         if(!info) throw std::runtime_error("there is no " + tag + " tag in INFO of this variant.\n");
         S * dst = NULL;
         ndst = 0;
         ret = -1;
         if(info->type == BCF_BT_INT8 || info->type == BCF_BT_INT16 || info->type == BCF_BT_INT32)
         {
-            ret = bcf_get_info_int32(header.hdr, line.get(), tag.c_str(), &dst, &ndst);
+            ret = bcf_get_info_int32(header->hdr, line.get(), tag.c_str(), &dst, &ndst);
         }
         else if(info->type == BCF_BT_FLOAT)
         {
-            ret = bcf_get_info_float(header.hdr, line.get(), tag.c_str(), &dst, &ndst);
+            ret = bcf_get_info_float(header->hdr, line.get(), tag.c_str(), &dst, &ndst);
         }
         if(ret >= 0)
+        {
             v = std::vector<S>(dst, dst + ret); // user have to check if there is missing in the return v;
+            free(dst);
+        }
         else
+        {
+            free(dst);
             throw std::runtime_error("couldn't parse the " + tag + " format of this variant.\n");
-        free(dst);
+        }
         return true;
     }
 
@@ -667,7 +702,7 @@ class BcfRecord
     template<typename T>
     isScalar<T> getINFO(std::string tag, T & v)
     {
-        info = bcf_get_info(header.hdr, line.get(), tag.c_str());
+        info = bcf_get_info(header->hdr, line.get(), tag.c_str());
         if(!info) throw std::runtime_error("there is no " + tag + " tag in INFO of this variant.\n");
         // scalar value
         if(info->len == 1)
@@ -697,7 +732,7 @@ class BcfRecord
     template<typename T>
     isString<T> getINFO(std::string tag, T & v)
     {
-        info = bcf_get_info(header.hdr, line.get(), tag.c_str());
+        info = bcf_get_info(header->hdr, line.get(), tag.c_str());
         if(!info) throw std::runtime_error("there is no " + tag + " tag in INFO of this variant.\n");
         if(info->type == BCF_BT_CHAR)
             v = std::string(reinterpret_cast<char *>(info->vptr), info->vptr_len);
@@ -717,13 +752,13 @@ class BcfRecord
         ret = -1;
         // bcf_hrec_set_val
         // bcf_update_info_flag(header.hdr, line, tag.c_str(), NULL, 1);
-        int tag_id = bcf_hdr_id2int(header.hdr, BCF_DT_ID, tag.c_str());
-        if(bcf_hdr_id2type(header.hdr, BCF_HL_INFO, tag_id) == (BCF_HT_INT & 0xff))
-            ret = bcf_update_info_int32(header.hdr, line.get(), tag.c_str(), &v, 1);
-        else if(bcf_hdr_id2type(header.hdr, BCF_HL_INFO, tag_id) == (BCF_HT_REAL & 0xff))
+        int tag_id = bcf_hdr_id2int(header->hdr, BCF_DT_ID, tag.c_str());
+        if(bcf_hdr_id2type(header->hdr, BCF_HL_INFO, tag_id) == (BCF_HT_INT & 0xff))
+            ret = bcf_update_info_int32(header->hdr, line.get(), tag.c_str(), &v, 1);
+        else if(bcf_hdr_id2type(header->hdr, BCF_HL_INFO, tag_id) == (BCF_HT_REAL & 0xff))
         {
             float v2 = static_cast<float>(v);
-            ret = bcf_update_info_float(header.hdr, line.get(), tag.c_str(), &v2, 1);
+            ret = bcf_update_info_float(header->hdr, line.get(), tag.c_str(), &v2, 1);
         }
         if(ret < 0)
         {
@@ -747,13 +782,13 @@ class BcfRecord
     {
         ret = -1;
         // bcf_update_info_flag(header.hdr, line, tag.c_str(), NULL, 1);
-        int tag_id = bcf_hdr_id2int(header.hdr, BCF_DT_ID, tag.c_str());
-        if(bcf_hdr_id2type(header.hdr, BCF_HL_INFO, tag_id) == (BCF_HT_INT & 0xff))
-            ret = bcf_update_info_int32(header.hdr, line.get(), tag.c_str(), v.data(), v.size());
-        else if(bcf_hdr_id2type(header.hdr, BCF_HL_INFO, tag_id) == (BCF_HT_REAL & 0xff))
-            ret = bcf_update_info_float(header.hdr, line.get(), tag.c_str(), v.data(), v.size());
-        else if(bcf_hdr_id2type(header.hdr, BCF_HL_INFO, tag_id) == (BCF_HT_STR & 0xff))
-            ret = bcf_update_info_string(header.hdr, line.get(), tag.c_str(), v.data());
+        int tag_id = bcf_hdr_id2int(header->hdr, BCF_DT_ID, tag.c_str());
+        if(bcf_hdr_id2type(header->hdr, BCF_HL_INFO, tag_id) == (BCF_HT_INT & 0xff))
+            ret = bcf_update_info_int32(header->hdr, line.get(), tag.c_str(), v.data(), v.size());
+        else if(bcf_hdr_id2type(header->hdr, BCF_HL_INFO, tag_id) == (BCF_HT_REAL & 0xff))
+            ret = bcf_update_info_float(header->hdr, line.get(), tag.c_str(), v.data(), v.size());
+        else if(bcf_hdr_id2type(header->hdr, BCF_HL_INFO, tag_id) == (BCF_HT_STR & 0xff))
+            ret = bcf_update_info_string(header->hdr, line.get(), tag.c_str(), v.data());
         if(ret < 0)
             throw std::runtime_error("couldn't set " + tag
                                      + " for this variant.\nplease add the tag in header first.\n");
@@ -765,13 +800,13 @@ class BcfRecord
     void removeINFO(std::string tag)
     {
         ret = -1;
-        int tag_id = bcf_hdr_id2int(header.hdr, BCF_DT_ID, tag.c_str());
-        if(bcf_hdr_id2type(header.hdr, BCF_HL_INFO, tag_id) == (BCF_HT_INT & 0xff))
-            ret = bcf_update_info_int32(header.hdr, line.get(), tag.c_str(), NULL, 0);
-        else if(bcf_hdr_id2type(header.hdr, BCF_HL_INFO, tag_id) == (BCF_HT_REAL & 0xff))
-            ret = bcf_update_info_float(header.hdr, line.get(), tag.c_str(), NULL, 0);
-        else if(bcf_hdr_id2type(header.hdr, BCF_HL_INFO, tag_id) == (BCF_HT_STR & 0xff))
-            ret = bcf_update_info_string(header.hdr, line.get(), tag.c_str(), NULL);
+        int tag_id = bcf_hdr_id2int(header->hdr, BCF_DT_ID, tag.c_str());
+        if(bcf_hdr_id2type(header->hdr, BCF_HL_INFO, tag_id) == (BCF_HT_INT & 0xff))
+            ret = bcf_update_info_int32(header->hdr, line.get(), tag.c_str(), NULL, 0);
+        else if(bcf_hdr_id2type(header->hdr, BCF_HL_INFO, tag_id) == (BCF_HT_REAL & 0xff))
+            ret = bcf_update_info_float(header->hdr, line.get(), tag.c_str(), NULL, 0);
+        else if(bcf_hdr_id2type(header->hdr, BCF_HL_INFO, tag_id) == (BCF_HT_STR & 0xff))
+            ret = bcf_update_info_string(header->hdr, line.get(), tag.c_str(), NULL);
         if(ret < 0)
         {
             throw std::runtime_error("couldn't remove " + tag + " for this variant.\n");
@@ -788,22 +823,26 @@ class BcfRecord
         // bcf_gt_type
         int i, j, k;
         nploidy = v.size() / nsamples;
-        gts = (int32_t *)malloc(v.size() * sizeof(int32_t));
+        int32_t * gt = (int32_t *)malloc(v.size() * sizeof(int32_t));
         for(i = 0; i < nsamples; i++)
         {
             for(j = 0; j < nploidy; j++)
             {
                 k = i * nploidy + j;
                 if(v[k] == -9 || v[k] == bcf_int32_missing)
-                    gts[k] = bcf_gt_missing;
+                    gt[k] = bcf_gt_missing;
                 else if(gtPhase[i])
-                    gts[k] = bcf_gt_phased(v[k]);
+                    gt[k] = bcf_gt_phased(v[k]);
                 else
-                    gts[k] = bcf_gt_unphased(v[k]);
+                    gt[k] = bcf_gt_unphased(v[k]);
             }
         }
-        if(bcf_update_genotypes(header.hdr, line.get(), gts, v.size()) < 0)
+        if(bcf_update_genotypes(header->hdr, line.get(), gt, v.size()) < 0)
+        {
+            free(gt);
             throw std::runtime_error("couldn't set genotypes correctly.\n");
+        }
+        free(gt);
         return true;
     }
 
@@ -821,13 +860,13 @@ class BcfRecord
     void removeFORMAT(std::string tag)
     {
         ret = -1;
-        int tag_id = bcf_hdr_id2int(header.hdr, BCF_DT_ID, tag.c_str());
-        if(bcf_hdr_id2type(header.hdr, BCF_HL_FMT, tag_id) == (BCF_HT_INT & 0xff))
-            ret = bcf_update_format_int32(header.hdr, line.get(), tag.c_str(), NULL, 0);
-        else if(bcf_hdr_id2type(header.hdr, BCF_HL_FMT, tag_id) == (BCF_HT_STR & 0xff))
-            ret = bcf_update_format_char(header.hdr, line.get(), tag.c_str(), NULL, 0);
-        else if(bcf_hdr_id2type(header.hdr, BCF_HL_FMT, tag_id) == (BCF_HT_REAL & 0xff))
-            ret = bcf_update_format_float(header.hdr, line.get(), tag.c_str(), NULL, 0);
+        int tag_id = bcf_hdr_id2int(header->hdr, BCF_DT_ID, tag.c_str());
+        if(bcf_hdr_id2type(header->hdr, BCF_HL_FMT, tag_id) == (BCF_HT_INT & 0xff))
+            ret = bcf_update_format_int32(header->hdr, line.get(), tag.c_str(), NULL, 0);
+        else if(bcf_hdr_id2type(header->hdr, BCF_HL_FMT, tag_id) == (BCF_HT_STR & 0xff))
+            ret = bcf_update_format_char(header->hdr, line.get(), tag.c_str(), NULL, 0);
+        else if(bcf_hdr_id2type(header->hdr, BCF_HL_FMT, tag_id) == (BCF_HT_REAL & 0xff))
+            ret = bcf_update_format_float(header->hdr, line.get(), tag.c_str(), NULL, 0);
         if(ret < 0) throw std::runtime_error("couldn't remove " + tag + " correctly.\n");
     }
 
@@ -841,13 +880,13 @@ class BcfRecord
     isValidFMT<T> setFORMAT(std::string tag, const T & v)
     {
         ret = -1;
-        int tag_id = bcf_hdr_id2int(header.hdr, BCF_DT_ID, tag.c_str());
-        if(bcf_hdr_id2type(header.hdr, BCF_HL_FMT, tag_id) == (BCF_HT_INT & 0xff))
-            ret = bcf_update_format_int32(header.hdr, line.get(), tag.c_str(), v.data(), v.size());
-        else if(bcf_hdr_id2type(header.hdr, BCF_HL_FMT, tag_id) == (BCF_HT_STR & 0xff))
-            ret = bcf_update_format_char(header.hdr, line.get(), tag.c_str(), v.data(), v.size());
-        else if(bcf_hdr_id2type(header.hdr, BCF_HL_FMT, tag_id) == (BCF_HT_REAL & 0xff))
-            ret = bcf_update_format_float(header.hdr, line.get(), tag.c_str(), v.data(), v.size());
+        int tag_id = bcf_hdr_id2int(header->hdr, BCF_DT_ID, tag.c_str());
+        if(bcf_hdr_id2type(header->hdr, BCF_HL_FMT, tag_id) == (BCF_HT_INT & 0xff))
+            ret = bcf_update_format_int32(header->hdr, line.get(), tag.c_str(), v.data(), v.size());
+        else if(bcf_hdr_id2type(header->hdr, BCF_HL_FMT, tag_id) == (BCF_HT_STR & 0xff))
+            ret = bcf_update_format_char(header->hdr, line.get(), tag.c_str(), v.data(), v.size());
+        else if(bcf_hdr_id2type(header->hdr, BCF_HL_FMT, tag_id) == (BCF_HT_REAL & 0xff))
+            ret = bcf_update_format_float(header->hdr, line.get(), tag.c_str(), v.data(), v.size());
         if(ret < 0) throw std::runtime_error("couldn't set format " + tag + " correctly.\n");
         return true;
     }
@@ -864,11 +903,11 @@ class BcfRecord
     {
         ret = -1;
         float v2 = v;
-        int tag_id = bcf_hdr_id2int(header.hdr, BCF_DT_ID, tag.c_str());
-        if(bcf_hdr_id2type(header.hdr, BCF_HL_FMT, tag_id) == (BCF_HT_INT & 0xff))
-            ret = bcf_update_format_int32(header.hdr, line.get(), tag.c_str(), &v, 1);
-        else if(bcf_hdr_id2type(header.hdr, BCF_HL_FMT, tag_id) == (BCF_HT_REAL & 0xff))
-            ret = bcf_update_format_float(header.hdr, line.get(), tag.c_str(), &v2, 1);
+        int tag_id = bcf_hdr_id2int(header->hdr, BCF_DT_ID, tag.c_str());
+        if(bcf_hdr_id2type(header->hdr, BCF_HL_FMT, tag_id) == (BCF_HT_INT & 0xff))
+            ret = bcf_update_format_int32(header->hdr, line.get(), tag.c_str(), &v, 1);
+        else if(bcf_hdr_id2type(header->hdr, BCF_HL_FMT, tag_id) == (BCF_HT_REAL & 0xff))
+            ret = bcf_update_format_float(header->hdr, line.get(), tag.c_str(), &v2, 1);
         if(ret < 0) throw std::runtime_error("couldn't set format " + tag + " correctly.\n");
         return true;
     }
@@ -876,23 +915,22 @@ class BcfRecord
     /** @brief add one variant record from given string*/
     void addLineFromString(const std::string & vcfline)
     {
-        std::vector<char> str(vcfline.begin(), vcfline.end());
-        str.push_back('\0'); // don't forget string has no \0;
-        kstring_t s = {vcfline.length(), vcfline.length(), &str[0]}; // kstring
-        ret = vcf_parse1(&s, header.hdr, line.get());
-        if(ret < 0) throw std::runtime_error("error parsing: " + vcfline + "\n");
+        kstring_t s = {0, 0, NULL};
+        kputsn(vcfline.c_str(), vcfline.length(), &s);
+        ret = vcf_parse1(&s, header->hdr, line.get());
+        free(s.s);
+        if(ret > 0) throw std::runtime_error("error parsing: " + vcfline + "\n");
         if(line->errcode == BCF_ERR_CTG_UNDEF)
         {
-            std::string contig(bcf_hdr_id2name(header.hdr, line->rid));
-            hdr_d = bcf_hdr_dup(header.hdr);
-            header.hrec = bcf_hdr_id2hrec(hdr_d, BCF_DT_CTG, 0, line->rid);
-            if(header.hrec == NULL)
+            std::string contig(bcf_hdr_id2name(header->hdr, line->rid));
+            hdr_d = bcf_hdr_dup(header->hdr);
+            header->hrec = bcf_hdr_id2hrec(hdr_d, BCF_DT_CTG, 0, line->rid);
+            if(header->hrec == NULL)
                 throw std::runtime_error("contig" + contig + " unknow and not found in the header.\n");
-            ret = bcf_hdr_add_hrec(header.hdr, header.hrec);
+            ret = bcf_hdr_add_hrec(header->hdr, header->hrec);
             if(ret < 0) throw std::runtime_error("error adding contig " + contig + " to header.\n");
-            ret = bcf_hdr_sync(header.hdr);
+            ret = bcf_hdr_sync(header->hdr);
         }
-        free(s.s);
     }
 
     /** @brief if all samples have non missing values for the tag in FORMAT */
@@ -904,7 +942,7 @@ class BcfRecord
     /** @brief return boolean value indicates if current variant is Structual Variant or not */
     inline bool isSV() const
     {
-        if(bcf_get_info(header.hdr, line.get(), "SVTYPE") == NULL)
+        if(bcf_get_info(header->hdr, line.get(), "SVTYPE") == NULL)
             return false;
         else
             return true;
@@ -1037,7 +1075,7 @@ class BcfRecord
     /** @brief return CHROM name */
     inline std::string CHROM() const
     {
-        return std::string(bcf_hdr_id2name(header.hdr, line->rid));
+        return std::string(bcf_hdr_id2name(header->hdr, line->rid));
     }
 
     /** @brief return ID field */
@@ -1055,7 +1093,7 @@ class BcfRecord
     /** @brief modify CHROM value */
     inline void setCHR(const char * chr)
     {
-        line->rid = bcf_hdr_name2id(header.hdr, chr);
+        line->rid = bcf_hdr_name2id(header->hdr, chr);
     }
 
     /** @brief modify position given 1-based value */
@@ -1067,13 +1105,13 @@ class BcfRecord
     /** @brief update ID */
     inline void setID(const char * s)
     {
-        bcf_update_id(header.hdr, line.get(), s);
+        bcf_update_id(header->hdr, line.get(), s);
     }
 
     /** @brief set REF and ALT alleles given a string seperated by comma */
     inline void setRefAlt(const char * alleles_string)
     {
-        bcf_update_alleles_str(header.hdr, line.get(), alleles_string);
+        bcf_update_alleles_str(header->hdr, line.get(), alleles_string);
     }
 
     /** @brief return 0-base start of the variant (can be any type) */
@@ -1135,14 +1173,14 @@ class BcfRecord
         }
         else if(line->d.n_flt == 1)
         {
-            return std::string(bcf_hdr_int2id(header.hdr, BCF_DT_ID, line->d.flt[0]));
+            return std::string(bcf_hdr_int2id(header->hdr, BCF_DT_ID, line->d.flt[0]));
         }
         else
         {
             std::string s;
             for(int i = 1; i < line->d.n_flt; i++)
             {
-                s += std::string(bcf_hdr_int2id(header.hdr, BCF_DT_ID, line->d.flt[i])) + ",";
+                s += std::string(bcf_hdr_int2id(header->hdr, BCF_DT_ID, line->d.flt[i])) + ",";
             }
             s.pop_back();
             return s;
@@ -1152,7 +1190,7 @@ class BcfRecord
     /** @brief return raw INFO column as string. recommend to use getINFO for specific tag. */
     inline std::string allINFO()
     {
-        int32_t max_dt_id = header.hdr->n[BCF_DT_ID];
+        int32_t max_dt_id = header->hdr->n[BCF_DT_ID];
         kstring_t * s = (kstring_t *)calloc(1, sizeof(kstring_t));
         if(line->n_info)
         {
@@ -1163,9 +1201,9 @@ class BcfRecord
                 if(!z->vptr) continue;
                 if(!first) kputc(';', s);
                 first = 0;
-                if(z->key < 0 || z->key >= max_dt_id || header.hdr->id[BCF_DT_ID][z->key].key == NULL)
+                if(z->key < 0 || z->key >= max_dt_id || header->hdr->id[BCF_DT_ID][z->key].key == NULL)
                     throw std::runtime_error("Invalid BCF and wrong INFO tag");
-                kputs(header.hdr->id[BCF_DT_ID][z->key].key, s);
+                kputs(header->hdr->id[BCF_DT_ID][z->key].key, s);
                 if(z->len <= 0) continue;
                 kputc('=', s);
                 if(z->len == 1)
@@ -1267,7 +1305,7 @@ class BcfReader
     std::shared_ptr<htsFile> fp; // hts file
     hts_idx_t * hidx = NULL; // hts index file
     tbx_t * tidx = NULL; // .tbi .csi index file for vcf files
-    hts_itr_t * itr = NULL; // hts records iterator
+    hts_itr_t * itr = NULL; // tabix iterator
     kstring_t s = {0, 0, NULL}; // kstring
     std::string fname;
     bool isBcf; // if the input file is bcf or vcf;
@@ -1321,10 +1359,19 @@ class BcfReader
         if(!samples.empty()) setSamples(samples);
     }
 
-    /// return a BcfHeader object
-    const BcfHeader & getHeader() const
+    ~BcfReader()
     {
-        return header;
+        close();
+    }
+
+    /// close the BcfReader object.
+    void close()
+    {
+        if(fp) fp.reset();
+        if(itr) hts_itr_destroy(itr);
+        if(hidx) hts_idx_destroy(hidx);
+        if(tidx) tbx_destroy(tidx);
+        if(s.s) free(s.s);
     }
 
     /// open a VCF/BCF/STDIN file for streaming in
@@ -1337,31 +1384,18 @@ class BcfReader
         SamplesName = header.getSamples();
     }
 
-    /// return if the file is opened successfully
-    bool isOpen() const
-    {
-        if(fp != NULL)
-            return true;
-        else
-            return false;
-    }
-
-    /// close the BcfReader object.
-    void close()
-    {
-        free(s.s);
-        if(fp) fp.reset();
-        if(itr) hts_itr_destroy(itr);
-    }
-
-    ~BcfReader() {}
-
     /** @brief set the number of threads to use */
     inline int setThreads(int n)
     {
         return hts_set_threads(fp.get(), n);
     }
 
+    /// return a BcfHeader object
+    const BcfHeader & getHeader() const
+    {
+        return header;
+    }
+
     /** @brief get the number of records of given region */
     uint64_t getVariantsCount(BcfRecord & r, const std::string & region)
     {
@@ -1434,15 +1468,15 @@ class BcfReader
                 int slen = tbx_itr_next(fp.get(), tidx, itr, &s);
                 if(slen > 0)
                 {
-                    ret = vcf_parse1(&s, r.header.hdr, r.line.get()); // ret < 0, error
+                    ret = vcf_parse1(&s, r.header->hdr, r.line.get()); // ret > 0, error
                     bcf_unpack(r.line.get(), BCF_UN_ALL);
                 }
-                return (ret >= 0) && (slen > 0);
+                return (ret <= 0) && (slen > 0);
             }
         }
         else
         {
-            ret = bcf_read(fp.get(), r.header.hdr, r.line.get());
+            ret = bcf_read(fp.get(), r.header->hdr, r.line.get());
             // unpack record immediately. not lazy
             bcf_unpack(r.line.get(), BCF_UN_ALL);
             return (ret == 0);
@@ -1461,6 +1495,7 @@ class BcfWriter
     std::shared_ptr<bcf1_t> b = std::shared_ptr<bcf1_t>(bcf_init(), variant_close()); // variant
     int ret;
     bool isHeaderWritten = false;
+    const BcfHeader * hp;
 
   public:
     /// header object initialized by initalHeader
@@ -1478,6 +1513,7 @@ class BcfWriter
     {
         open(fname);
         initalHeader(version);
+        initalHeader(header);
     }
 
     /**
@@ -1505,6 +1541,7 @@ class BcfWriter
     {
         open(fname, mode);
         initalHeader(version);
+        initalHeader(header);
     }
 
     /**
@@ -1531,10 +1568,7 @@ class BcfWriter
      */
     void open(const std::string & fname)
     {
-        std::string mode{"w"};
-        if(isEndWith(fname, "bcf.gz")) mode += "b";
-        if(isEndWith(fname, "bcf")) mode += "bu";
-        if(isEndWith(fname, "vcf.gz")) mode += "z";
+        auto mode = getMode(fname, "w");
         fp = std::shared_ptr<htsFile>(hts_open(fname.c_str(), mode.c_str()), htsFile_close());
     }
 
@@ -1567,36 +1601,43 @@ class BcfWriter
         header.setVersion(version);
     }
 
-    /// initial a VCF header by refering to another vcf header
+    /// initial a VCF header by pointing to header of another VCF
     void initalHeader(const BcfHeader & h)
     {
-        header.hdr = h.hdr; // point to another header
-        if(header.hdr == NULL) throw std::runtime_error("couldn't copy the header from another vcf.\n");
+        hp = &h;
     }
 
-    /// write a string to a vcf line
+    /// copy header of given VCF
+    void copyHeader(const std::string & vcffile)
+    {
+        htsFile * fp2 = hts_open(vcffile.c_str(), "r");
+        header.hdr = bcf_hdr_read(fp2);
+        hts_close(fp2);
+        initalHeader(header);
+    }
+
+    /// copy a string to a vcf line
     void writeLine(const std::string & vcfline)
     {
         if(!isHeaderWritten && !writeHeader()) throw std::runtime_error("could not write header\n");
-        std::vector<char> str(vcfline.begin(), vcfline.end());
-        str.push_back('\0'); // don't forget string has no \0;
-        kstring_t s = {vcfline.length(), vcfline.length(), &str[0]}; // kstring
-        ret = vcf_parse1(&s, header.hdr, b.get());
-        if(ret < 0) hts_log_debug("the ret is %i ", ret);
-        if(ret < 0) throw std::runtime_error("error parsing: " + vcfline + "\n");
+        kstring_t s = {0, 0, NULL};
+        kputsn(vcfline.c_str(), vcfline.length(), &s);
+        ret = vcf_parse1(&s, hp->hdr, b.get());
+        free(s.s);
+        if(ret > 0) throw std::runtime_error("error parsing: " + vcfline + "\n");
         if(b->errcode == BCF_ERR_CTG_UNDEF)
         {
-            throw std::runtime_error("contig id " + (std::string)bcf_hdr_id2name(header.hdr, b->rid)
+            throw std::runtime_error("contig id " + (std::string)bcf_hdr_id2name(hp->hdr, b->rid)
                                      + " not found in the header. please run header->AddContig() first.\n");
         }
-        ret = bcf_write(fp.get(), header.hdr, b.get());
+        ret = bcf_write(fp.get(), hp->hdr, b.get());
         if(ret != 0) throw std::runtime_error("error writing: " + vcfline + "\n");
     }
 
     /// streams out the header
     bool writeHeader()
     {
-        ret = bcf_hdr_write(fp.get(), header.hdr);
+        ret = bcf_hdr_write(fp.get(), hp->hdr);
         if(ret == 0)
             return isHeaderWritten = true;
         else
@@ -1607,7 +1648,7 @@ class BcfWriter
     inline bool writeRecord(BcfRecord & v)
     {
         if(!isHeaderWritten) writeHeader();
-        if(bcf_write(fp.get(), v.header.hdr, v.line.get()) < 0)
+        if(bcf_write(fp.get(), v.header->hdr, v.line.get()) < 0)
             return false;
         else
             return true;
